@@ -3,12 +3,12 @@
    Find and identify the network interfaces. */
 
 /*
- * Copyright (c) 2004-2016 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2017 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
  * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
@@ -44,6 +44,7 @@ int interfaces_invalidated;
 int quiet_interface_discovery;
 u_int16_t local_port;
 u_int16_t remote_port;
+u_int16_t relay_port = 0;
 int dhcpv4_over_dhcpv6 = 0;
 int (*dhcp_interface_setup_hook) (struct interface_info *, struct iaddr *);
 int (*dhcp_interface_discovery_hook) (struct interface_info *);
@@ -55,6 +56,14 @@ struct in_addr limited_broadcast;
 int local_family = AF_INET;
 struct in_addr local_address;
 
+#ifdef DHCPv6
+/*
+ * Another clear abuse of the fact that undefined IP addresses are all zeroes.
+ */
+struct in6_addr local_address6;
+int bind_local_address6 = 0;
+#endif /* DHCPv6 */
+
 void (*bootp_packet_handler) (struct interface_info *,
 			      struct dhcp_packet *, unsigned,
 			      unsigned int,
@@ -376,10 +385,10 @@ end_iface_scan(struct iface_conf_list *ifaces) {
 #else
 
 /* 
- * Unix support
+ * BSD/Linux support
  * -----------
  *
- * FreeBSD, NetBSD, OpenBSD, Linux, and OS X all have the getifaddrs()
+ * FreeBSD, NetBSD, OpenBSD, OS X/macOS and Linux all have the getifaddrs() 
  * function.
  *
  * The getifaddrs() man page describes the use.
@@ -441,17 +450,20 @@ next_iface(struct iface_info *info, int *err, struct iface_conf_list *ifaces) {
 	}
 	memset(info, 0, sizeof(struct iface_info));
 	strncpy(info->name, ifaces->next->ifa_name, sizeof(info->name) - 1);
-
 	memset(&info->addr, 0 , sizeof(info->addr));
-
+	/*
+	 * getifaddrs() can on Linux with some interfaces like PPP or TEQL
+	 * result in a record with no address (ifa_addr).
+	 */
 	if (ifaces->next->ifa_addr != NULL) {
-#ifdef HAVE_SA_LEN
-		sa_len = ifaces->next->ifa_addr->sa_len;
-#else
+/* Linux lacks the sa_len member in struct sockaddr. */
+#if defined(__linux)
 		if (ifaces->next->ifa_addr->sa_family == AF_INET)
 			sa_len = sizeof(struct sockaddr_in);
 		else if (ifaces->next->ifa_addr->sa_family == AF_INET6)
 			sa_len = sizeof(struct sockaddr_in6);
+#else
+		sa_len = ifaces->next->ifa_addr->sa_len;
 #endif
 		memcpy(&info->addr, ifaces->next->ifa_addr, sa_len);
 	}
@@ -570,6 +582,10 @@ discover_interfaces(int state) {
 	int ir;
 	isc_result_t status;
 	int wifcount = 0;
+#ifdef RELAY_PORT
+	int updone = 0;
+	int downdone = 0;
+#endif
 
 	static int setup_fallback = 0;
 
@@ -935,9 +951,39 @@ discover_interfaces(int state) {
 		switch (local_family) {
 #ifdef DHCPv6 
 		case AF_INET6:
+#ifdef RELAY_PORT
+#define UPSTREAM(ifp) \
+	((ifp->flags & INTERFACE_STREAMS) == INTERFACE_UPSTREAM)
+#define DOWNSTREAM(ifp) \
+        ((ifp->flags & INTERFACE_STREAMS) == INTERFACE_DOWNSTREAM)
+
+			if (relay_port) {
+				/*
+				 * The normal IPv6 relay only needs one
+				 * socket as long as we find an interface.
+				 * When user relay port is defined, and we
+				 * have two different UDP ports. One to
+				 * receive from DHCP client with port 547,
+				 * and the other is user defined for sending
+				 * to the server or upstream relay agent.
+				 * Thus we need to register sockets for one
+				 * upstream and one downstream interfaces.
+				 */
+				if (updone && UPSTREAM(tmp))
+					    continue;
+				if (downdone && DOWNSTREAM(tmp))
+					    continue;
+			}
+#endif
 			status = omapi_register_io_object((omapi_object_t *)tmp,
 							  if_readsocket, 
 							  0, got_one_v6, 0, 0);
+#ifdef RELAY_PORT
+			if (UPSTREAM(tmp))
+				updone++;
+			else
+				downdone++;
+#endif
 			break;
 #endif /* DHCPv6 */
 		case AF_INET:
@@ -959,8 +1005,12 @@ discover_interfaces(int state) {
 		 * dynamically adding and removing interfaces, but
 		 * we're well beyond that point in terms of mess.
 		 */
-		if (((state == DISCOVER_SERVER) || (state == DISCOVER_RELAY)) &&
-		    (local_family == AF_INET6))
+		if (((state == DISCOVER_SERVER) || (state == DISCOVER_RELAY))
+		    && (local_family == AF_INET6)
+#if defined(RELAY_PORT)
+		    && ((relay_port == 0) || (updone && downdone))
+#endif
+		    )
 			break;
 #endif
 	} /* for (tmp = interfaces; ... */
@@ -1488,8 +1538,11 @@ void interface_stash (struct interface_info *tptr)
 		delta = tptr -> index - interface_max + 10;
 		vec = dmalloc ((interface_max + delta) *
 			       sizeof (struct interface_info *), MDL);
-		if (!vec)
+		if (!vec) {
+			log_error ("interface_stash: allocation failed ");
 			return;
+		}
+
 		memset (&vec [interface_max], 0,
 			(sizeof (struct interface_info *)) * delta);
 		interface_max += delta;
@@ -1499,8 +1552,10 @@ void interface_stash (struct interface_info *tptr)
 			     sizeof (struct interface_info *)));
 		    dfree (interface_vector, MDL);
 		}
+
 		interface_vector = vec;
 	}
+
 	interface_reference (&interface_vector [tptr -> index], tptr, MDL);
 	if (tptr -> index >= interface_count)
 		interface_count = tptr -> index + 1;
