@@ -3,12 +3,12 @@
    Parser for dhclient config and lease files... */
 
 /*
- * Copyright (c) 2004-2014 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2017 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
  * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
@@ -32,7 +32,8 @@
 struct client_config top_level_config;
 
 #define NUM_DEFAULT_REQUESTED_OPTS	9
-struct option *default_requested_options[NUM_DEFAULT_REQUESTED_OPTS + 1];
+/* There can be 2 extra requested options for DHCPv4-over-DHCPv6. */
+struct option *default_requested_options[NUM_DEFAULT_REQUESTED_OPTS + 2 + 1];
 
 static void parse_client_default_duid(struct parse *cfile);
 static void parse_client6_lease_statement(struct parse *cfile);
@@ -44,6 +45,8 @@ static struct dhc6_addr *parse_client6_iaaddr_statement(struct parse *cfile);
 static struct dhc6_addr *parse_client6_iaprefix_statement(struct parse *cfile);
 #endif /* DHCPv6 */
 
+static void parse_lease_id_format (struct parse *cfile);
+
 /* client-conf-file :== client-declarations END_OF_FILE
    client-declarations :== <nil>
 			 | client-declaration
@@ -120,6 +123,43 @@ isc_result_t read_client_conf ()
 				  "assembly.", code);
 	}
 
+#ifdef DHCP4o6
+	/* DHCPv4-over-DHCPv6 extra requested options in code order */
+	if (dhcpv4_over_dhcpv6 == 1) {
+		/* The DHCP4o6 server option should be requested */
+		code = D6O_DHCP4_O_DHCP6_SERVER;
+		option_code_hash_lookup(&default_requested_options[9],
+					dhcpv6_universe.code_hash,
+					&code, 0, MDL);
+		if (default_requested_options[9] == NULL) {
+			log_fatal("Unable to find option definition for "
+				  "index %u during default parameter request "
+				  "assembly.", code);
+		}
+	} else if (dhcpv4_over_dhcpv6 > 1) {
+		/* Called from run_stateless so the IRT should
+		   be requested too */
+		code = D6O_INFORMATION_REFRESH_TIME;
+		option_code_hash_lookup(&default_requested_options[9],
+					dhcpv6_universe.code_hash,
+					&code, 0, MDL);
+		if (default_requested_options[9] == NULL) {
+			log_fatal("Unable to find option definition for "
+				  "index %u during default parameter request "
+				  "assembly.", code);
+		}
+		code = D6O_DHCP4_O_DHCP6_SERVER;
+		option_code_hash_lookup(&default_requested_options[10],
+					dhcpv6_universe.code_hash,
+					&code, 0, MDL);
+		if (default_requested_options[10] == NULL) {
+			log_fatal("Unable to find option definition for "
+				  "index %u during default parameter request "
+				  "assembly.", code);
+		}
+	}
+#endif
+					
 	/* Initialize the top level client configuration. */
 	memset (&top_level_config, 0, sizeof top_level_config);
 
@@ -130,6 +170,7 @@ isc_result_t read_client_conf ()
 	top_level_config.retry_interval = 300;
 	top_level_config.backoff_cutoff = 15;
 	top_level_config.initial_interval = 3;
+	top_level_config.lease_id_format = TOKEN_OCTAL;
 
 	/*
 	 * RFC 2131, section 4.4.1 specifies that the client SHOULD wait a
@@ -773,6 +814,12 @@ void parse_client_statement (cfile, ip, config)
 		parse_reject_statement (cfile, config);
 		return;
 
+	      case LEASE_ID_FORMAT:
+		skip_token(&val, (unsigned *)0, cfile);
+		parse_lease_id_format(cfile);
+		break;
+
+
 	      case BOOTP_BROADCAST_ALWAYS:
 		token = next_token(&val, (unsigned*)0, cfile);
 		config -> bootp_broadcast_always = 1;
@@ -1294,25 +1341,17 @@ static void
 parse_client_default_duid(struct parse *cfile)
 {
 	struct data_string new_duid;
-	const char *val = NULL;
+	u_int8_t buf[128];
 	unsigned len;
-	int token;
-
-	memset(&new_duid, 0, sizeof(new_duid));
-
-	token = next_token(&val, &len, cfile);
-	if (token != STRING) {
-		parse_warn(cfile, "Expected DUID string.");
-		skip_to_semi(cfile);
-		return;
-	}
 
+	len = parse_X(cfile, buf, sizeof(buf));
 	if (len <= 2) {
 		parse_warn(cfile, "Invalid DUID contents.");
 		skip_to_semi(cfile);
 		return;
 	}
 
+	memset(&new_duid, 0, sizeof(new_duid));
 	if (!buffer_allocate(&new_duid.buffer, len, MDL)) {
 		parse_warn(cfile, "Out of memory parsing default DUID.");
 		skip_to_semi(cfile);
@@ -1321,7 +1360,7 @@ parse_client_default_duid(struct parse *cfile)
 	new_duid.data = new_duid.buffer->data;
 	new_duid.len = len;
 
-	memcpy(new_duid.buffer->data, val, len);
+	memcpy(new_duid.buffer->data, buf, len);
 
 	/* Rotate the last entry into place. */
 	if (default_duid.buffer != NULL)
@@ -2280,3 +2319,45 @@ int parse_allow_deny (oc, cfile, flag)
 	skip_to_semi (cfile);
 	return 0;
 }
+
+
+
+/*!
+ * \brief Parses an lease-id-format statement
+ *
+ * A valid statement looks like this:
+ *
+ *	lease-id-format :==
+ *		LEASE_ID_FORMAT TOKEN_OCTAL | TOKEN_HEX ;
+ *
+ * This function is used to parse the lease-id-format statement. It sets
+ * top_level_config.lease_id_format.
+ *
+ * \param cfile the current parse file
+ *
+*/
+void parse_lease_id_format (struct parse *cfile)
+{
+	enum dhcp_token token;
+	const char *val;
+
+	token = next_token(&val, NULL, cfile);
+	switch(token) {
+	case TOKEN_OCTAL:
+		top_level_config.lease_id_format = TOKEN_OCTAL;
+		break;
+	case TOKEN_HEX:
+		top_level_config.lease_id_format = TOKEN_HEX;
+		break;
+	default:
+		parse_warn(cfile, "lease-id-format is invalid: "
+                                   " it must be octal or hex.");
+		skip_to_semi(cfile);
+		return;
+	}
+
+	log_debug("lease_id_format is: %s",
+		  (top_level_config.lease_id_format == TOKEN_OCTAL
+		   ? "octal" : "hex"));
+
+}
