@@ -3,13 +3,12 @@
    Find and identify the network interfaces. */
 
 /*
- * Copyright (c) 2013-2014 by Internet Systems Consortium, Inc. ("ISC")
- * Copyright (c) 2004-2009,2011 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2017 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
  * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
@@ -45,6 +44,8 @@ int interfaces_invalidated;
 int quiet_interface_discovery;
 u_int16_t local_port;
 u_int16_t remote_port;
+u_int16_t relay_port = 0;
+int dhcpv4_over_dhcpv6 = 0;
 int (*dhcp_interface_setup_hook) (struct interface_info *, struct iaddr *);
 int (*dhcp_interface_discovery_hook) (struct interface_info *);
 isc_result_t (*dhcp_interface_startup_hook) (struct interface_info *);
@@ -55,6 +56,14 @@ struct in_addr limited_broadcast;
 int local_family = AF_INET;
 struct in_addr local_address;
 
+#ifdef DHCPv6
+/*
+ * Another clear abuse of the fact that undefined IP addresses are all zeroes.
+ */
+struct in6_addr local_address6;
+int bind_local_address6 = 0;
+#endif /* DHCPv6 */
+
 void (*bootp_packet_handler) (struct interface_info *,
 			      struct dhcp_packet *, unsigned,
 			      unsigned int,
@@ -332,8 +341,8 @@ next_iface(struct iface_info *info, int *err, struct iface_conf_list *ifaces) {
 			continue;
 		}
 
-		strcpy(info->name, p->lifr_name);
-		memset(&info->addr, 0, sizeof(info->addr));
+		memset(info, 0, sizeof(struct iface_info));
+		strncpy(info->name, p->lifr_name, sizeof(info->name) - 1);
 		memcpy(&info->addr, &p->lifr_addr, sizeof(p->lifr_addr));
 
 #if defined(sun) || defined(__linux)
@@ -349,7 +358,7 @@ next_iface(struct iface_info *info, int *err, struct iface_conf_list *ifaces) {
 		 (strncmp(info->name, "dummy", 5) == 0));
 	
 	memset(&tmp, 0, sizeof(tmp));
-	strcpy(tmp.lifr_name, info->name);
+	strncpy(tmp.lifr_name, info->name, sizeof(tmp.lifr_name) - 1);
 	if (ioctl(ifaces->sock, SIOCGLIFFLAGS, &tmp) < 0) {
 		log_error("Error getting interface flags for '%s'; %m", 
 			  p->lifr_name);
@@ -376,10 +385,10 @@ end_iface_scan(struct iface_conf_list *ifaces) {
 #else
 
 /* 
- * BSD support
+ * BSD/Linux support
  * -----------
  *
- * FreeBSD, NetBSD, OpenBSD, OS X and Linux all have the getifaddrs()
+ * FreeBSD, NetBSD, OpenBSD, OS X/macOS and Linux all have the getifaddrs() 
  * function.
  *
  * The getifaddrs() man page describes the use.
@@ -439,18 +448,22 @@ next_iface(struct iface_info *info, int *err, struct iface_conf_list *ifaces) {
 		*err = 1;
 		return 0;
 	}
-	strcpy(info->name, ifaces->next->ifa_name);
-
+	memset(info, 0, sizeof(struct iface_info));
+	strncpy(info->name, ifaces->next->ifa_name, sizeof(info->name) - 1);
 	memset(&info->addr, 0 , sizeof(info->addr));
-
+	/*
+	 * getifaddrs() can on Linux with some interfaces like PPP or TEQL
+	 * result in a record with no address (ifa_addr).
+	 */
 	if (ifaces->next->ifa_addr != NULL) {
-#ifdef HAVE_SA_LEN
-		sa_len = ifaces->next->ifa_addr->sa_len;
-#else
+/* Linux lacks the sa_len member in struct sockaddr. */
+#if defined(__linux)
 		if (ifaces->next->ifa_addr->sa_family == AF_INET)
 			sa_len = sizeof(struct sockaddr_in);
 		else if (ifaces->next->ifa_addr->sa_family == AF_INET6)
 			sa_len = sizeof(struct sockaddr_in6);
+#else
+		sa_len = ifaces->next->ifa_addr->sa_len;
 #endif
 		memcpy(&info->addr, ifaces->next->ifa_addr, sa_len);
 	}
@@ -569,6 +582,10 @@ discover_interfaces(int state) {
 	int ir;
 	isc_result_t status;
 	int wifcount = 0;
+#ifdef RELAY_PORT
+	int updone = 0;
+	int downdone = 0;
+#endif
 
 	static int setup_fallback = 0;
 
@@ -584,11 +601,18 @@ discover_interfaces(int state) {
 		ir = 0;
 	else if (state == DISCOVER_UNCONFIGURED)
 		ir = INTERFACE_REQUESTED | INTERFACE_AUTOMATIC;
-	else
+	else {
 		ir = INTERFACE_REQUESTED;
+		if (state == DISCOVER_RELAY && local_family == AF_INET) {
+			/* We're a v4 relay without specifically requested
+			 * interfaces, so mark them all as bidirectional. */
+			ir |= INTERFACE_STREAMS;
+		}
+	}
 
 	/* Cycle through the list of interfaces looking for IP addresses. */
 	while (next_iface(&info, &err, &ifaces)) {
+		struct ifaddrs *ifaddr, *ifa;
 
 		/* See if we've seen an interface that matches this one. */
 		for (tmp = interfaces; tmp; tmp = tmp->next) {
@@ -620,7 +644,7 @@ discover_interfaces(int state) {
 				log_fatal("Error allocating interface %s: %s",
 					  info.name, isc_result_totext(status));
 			}
-			strcpy(tmp->name, info.name);
+			strncpy(tmp->name, info.name, sizeof(tmp->name) - 1);
 			interface_snorf(tmp, ir);
 			interface_dereference(&tmp, MDL);
 			tmp = interfaces; /* XXX */
@@ -630,31 +654,33 @@ discover_interfaces(int state) {
 			(*dhcp_interface_discovery_hook)(tmp);
 		}
 
-		{
-		  struct ifaddrs *ifaddr, *ifa;
-		  getifaddrs(&ifaddr);
-		  for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
-		    if ( (ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_INET) && 
-			 (local_family == AF_INET)) {
-		      struct sockaddr_in *a = (struct sockaddr_in *)ifa->ifa_addr;
-		      struct iaddr addr;
-		      char *s=NULL;
+		getifaddrs(&ifaddr);
+		for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
 
-		      s = strchr(ifa->ifa_name, ':');
-		      if (s != NULL) {
+		    char *s=NULL;
+
+		    s = strchr(ifa->ifa_name, ':');
+		    if (s != NULL) {
 			*s = '\0';
-		      }
-		      if (!strcmp(ifa->ifa_name, info.name)) {
+		    }
+		    if (strcmp(ifa->ifa_name, info.name))
+			continue;
+		    if ( (ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_INET) &&
+		         (local_family == AF_INET)) {
+			struct sockaddr_in *a = (struct sockaddr_in *)ifa->ifa_addr;
+			struct iaddr addr;
+
 			/* We don't want the loopback interface. */
 			if (a->sin_addr.s_addr == htonl(INADDR_LOOPBACK) &&
 			    ((tmp->flags & INTERFACE_AUTOMATIC) &&
-			     state == DISCOVER_SERVER))
-			  continue;
-			
+			     ((state == DISCOVER_SERVER) ||
+			      (state == DISCOVER_SERVER46))))
+				continue;
+
 			/* If the only address we have is 0.0.0.0, we
 			   shouldn't consider the interface configured. */
 			if (a->sin_addr.s_addr != htonl(INADDR_ANY))
-			  tmp->configured = 1;
+				tmp->configured = 1;
 
 			add_ipv4_addr_to_interface(tmp, &a->sin_addr);
 
@@ -662,46 +688,38 @@ discover_interfaces(int state) {
 			addr.len = 4;
 			memcpy(addr.iabuf, &a->sin_addr.s_addr, addr.len);
 			if (dhcp_interface_setup_hook) {
-			  (*dhcp_interface_setup_hook)(tmp, &addr);
+				(*dhcp_interface_setup_hook)(tmp, &addr);
 			}
-		      }	 
-		    } 
+		    }
 #ifdef DHCPv6
-		    else if ( (ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_INET6) && 
-			      (local_family == AF_INET6)) {
-		      struct sockaddr_in6 *a = 
-			(struct sockaddr_in6*)ifa->ifa_addr;
-		      struct iaddr addr;
-		      char *s=NULL;
+		    else if ((ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_INET6) &&
+			     (local_family == AF_INET6)) {
+			struct sockaddr_in6 *a =
+			              (struct sockaddr_in6*)ifa->ifa_addr;
+			struct iaddr addr;
 
-		      s = strchr(ifa->ifa_name, ':');
-		      if (s != NULL) {
-			*s = '\0';
-		      }		      
-		      if (!strcmp(ifa->ifa_name, info.name)) {
 			/* We don't want the loopback interface. */
 			if (IN6_IS_ADDR_LOOPBACK(&a->sin6_addr) && 
 			    ((tmp->flags & INTERFACE_AUTOMATIC) &&
-			     state == DISCOVER_SERVER))
-			  continue;
-			
+			     ((state == DISCOVER_SERVER) ||
+			      (state == DISCOVER_SERVER46))))
+			    continue;
+
 			/* If the only address we have is 0.0.0.0, we
 			   shouldn't consider the interface configured. */
 			if (IN6_IS_ADDR_UNSPECIFIED(&a->sin6_addr))
-			  tmp->configured = 1;
-			
+				tmp->configured = 1;
+
 			add_ipv6_addr_to_interface(tmp, &a->sin6_addr);
-			
+
 			/* invoke the setup hook */
 			addr.len = 16;
 			memcpy(addr.iabuf, &a->sin6_addr, addr.len);
 			if (dhcp_interface_setup_hook) {
-			  (*dhcp_interface_setup_hook)(tmp, &addr);
+				(*dhcp_interface_setup_hook)(tmp, &addr);
 			}
-		      }
 		    }
 #endif /* DHCPv6 */
-		  }
 		}
 	}
 
@@ -758,9 +776,9 @@ discover_interfaces(int state) {
 					  INTERFACE_REQUESTED);
 
 #ifdef DHCPv6
-		if (!(tmp->flags & INTERFACE_REQUESTED)) {
+		if (!(tmp->flags & INTERFACE_REQUESTED) && state != DISCOVER_RELAY) {
 #else
-		if (!tmp -> ifp || !(tmp -> flags & INTERFACE_REQUESTED)) {
+		if (!tmp -> ifp || (!(tmp -> flags & INTERFACE_REQUESTED) && state != DISCOVER_RELAY)) {
 #endif /* DHCPv6 */
 			if ((tmp -> flags & INTERFACE_REQUESTED) != ir)
 				log_fatal ("%s: not found", tmp -> name);
@@ -882,31 +900,48 @@ discover_interfaces(int state) {
 		tmp -> index = -1;
 
 		/* Register the interface... */
-		if (local_family == AF_INET) {
-			if_register_receive(tmp);
-			if_register_send(tmp);
+		switch (local_family) {
+		case AF_INET:
+			if (!dhcpv4_over_dhcpv6) {
+				if_register_receive(tmp);
+				if_register_send(tmp);
+			} else {
+				/* get_hw_addr() was called by register. */
+				get_hw_addr(tmp);
+			}
+			break;
 #ifdef DHCPv6
-		} else {
+		case AF_INET6:
 			if ((state == DISCOVER_SERVER) ||
 			    (state == DISCOVER_RELAY)) {
 				if_register6(tmp, 1);
+			} else if (state == DISCOVER_SERVER46) {
+				/* get_hw_addr() was called by if_register*6
+				   so now we have to call it explicitly
+				   to not leave the hardware address unknown
+				   (some code expects it cannot be. */
+				get_hw_addr(tmp);
 			} else {
 				if_register_linklocal6(tmp);
 			}
+			break;
 #endif /* DHCPv6 */
 		}
 
 		interface_stash (tmp);
 		wifcount++;
 #if defined (F_SETFD)
-		if (fcntl (tmp -> rfdesc, F_SETFD, 1) < 0)
+		/* if_register*() are no longer always called so
+		   descriptors  must be checked. */
+		if ((tmp -> rfdesc >= 0) &&
+		    (fcntl (tmp -> rfdesc, F_SETFD, 1) < 0))
+			log_error ("Can't set close-on-exec on %s: %m",
+				   tmp -> name);
+		if ((tmp -> wfdesc != tmp -> rfdesc) &&
+		    (tmp -> wfdesc >= 0) &&
+		    (fcntl (tmp -> wfdesc, F_SETFD, 1) < 0))
 			log_error ("Can't set close-on-exec on %s: %m",
 				   tmp -> name);
-		if (tmp -> rfdesc != tmp -> wfdesc) {
-			if (fcntl (tmp -> wfdesc, F_SETFD, 1) < 0)
-				log_error ("Can't set close-on-exec on %s: %m",
-					   tmp -> name);
-		}
 #endif
 	      next:
 		interface_dereference (&tmp, MDL);
@@ -929,9 +964,39 @@ discover_interfaces(int state) {
 		switch (local_family) {
 #ifdef DHCPv6 
 		case AF_INET6:
+#ifdef RELAY_PORT
+#define UPSTREAM(ifp) \
+	((ifp->flags & INTERFACE_STREAMS) == INTERFACE_UPSTREAM)
+#define DOWNSTREAM(ifp) \
+        ((ifp->flags & INTERFACE_STREAMS) == INTERFACE_DOWNSTREAM)
+
+			if (relay_port) {
+				/*
+				 * The normal IPv6 relay only needs one
+				 * socket as long as we find an interface.
+				 * When user relay port is defined, and we
+				 * have two different UDP ports. One to
+				 * receive from DHCP client with port 547,
+				 * and the other is user defined for sending
+				 * to the server or upstream relay agent.
+				 * Thus we need to register sockets for one
+				 * upstream and one downstream interfaces.
+				 */
+				if (updone && UPSTREAM(tmp))
+					    continue;
+				if (downdone && DOWNSTREAM(tmp))
+					    continue;
+			}
+#endif
 			status = omapi_register_io_object((omapi_object_t *)tmp,
 							  if_readsocket, 
 							  0, got_one_v6, 0, 0);
+#ifdef RELAY_PORT
+			if (UPSTREAM(tmp))
+				updone++;
+			else
+				downdone++;
+#endif
 			break;
 #endif /* DHCPv6 */
 		case AF_INET:
@@ -953,8 +1018,12 @@ discover_interfaces(int state) {
 		 * dynamically adding and removing interfaces, but
 		 * we're well beyond that point in terms of mess.
 		 */
-		if (((state == DISCOVER_SERVER) || (state == DISCOVER_RELAY)) &&
-		    (local_family == AF_INET6))
+		if (((state == DISCOVER_SERVER) || (state == DISCOVER_RELAY))
+		    && (local_family == AF_INET6)
+#if defined(RELAY_PORT)
+		    && ((relay_port == 0) || (updone && downdone))
+#endif
+		    )
 			break;
 #endif
 	} /* for (tmp = interfaces; ... */
@@ -964,7 +1033,8 @@ discover_interfaces(int state) {
 		log_fatal ("Not configured to listen on any interfaces!");
 	}
 
-	if ((local_family == AF_INET) && !setup_fallback) {
+	if ((local_family == AF_INET) &&
+	    !setup_fallback && !dhcpv4_over_dhcpv6) {
 		setup_fallback = 1;
 		maybe_setup_fallback();
 	}
@@ -1481,8 +1551,11 @@ void interface_stash (struct interface_info *tptr)
 		delta = tptr -> index - interface_max + 10;
 		vec = dmalloc ((interface_max + delta) *
 			       sizeof (struct interface_info *), MDL);
-		if (!vec)
+		if (!vec) {
+			log_error ("interface_stash: allocation failed ");
 			return;
+		}
+
 		memset (&vec [interface_max], 0,
 			(sizeof (struct interface_info *)) * delta);
 		interface_max += delta;
@@ -1492,8 +1565,10 @@ void interface_stash (struct interface_info *tptr)
 			     sizeof (struct interface_info *)));
 		    dfree (interface_vector, MDL);
 		}
+
 		interface_vector = vec;
 	}
+
 	interface_reference (&interface_vector [tptr -> index], tptr, MDL);
 	if (tptr -> index >= interface_count)
 		interface_count = tptr -> index + 1;
