@@ -3,12 +3,12 @@
    Tokens for config file lexer and parser. */
 
 /*
- * Copyright (c) 2004-2016 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2017 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
  * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
@@ -39,6 +39,8 @@ enum dhcp_token {
 	SLASH = '/',
 	LBRACE = '{',
 	RBRACE = '}',
+	LBRACKET = '[',
+	RBRACKET = ']',
 	LPAREN = '(',
 	RPAREN = ')',
 	EQUAL = '=',
@@ -374,7 +376,8 @@ enum dhcp_token {
 	LEASE_ID_FORMAT = 676,
 	TOKEN_HEX = 677,
 	TOKEN_OCTAL = 678,
-	BOOTP_BROADCAST_ALWAYS = 679
+	KEY_ALGORITHM = 679,
+	BOOTP_BROADCAST_ALWAYS = 680
 };
 
 #define is_identifier(x)	((x) >= FIRST_TOKEN &&	\
