@@ -3,12 +3,12 @@
    DHCP/BOOTP Relay Agent. */
 
 /*
- * Copyright(c) 2004-2016 by Internet Systems Consortium, Inc.("ISC")
+ * Copyright(c) 2004-2018 by Internet Systems Consortium, Inc.("ISC")
  * Copyright(c) 1997-2003 by Internet Software Consortium
  *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
  * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
@@ -77,6 +77,9 @@ int missing_circuit_id = 0;	/* Circuit ID option in matching RAI option
 				   was missing. */
 int max_hop_count = 10;		/* Maximum hop count */
 
+int no_daemon = 0;
+int dfd[2] = { -1, -1 };
+
 #ifdef DHCPv6
 	/* Force use of DHCPv6 interface-id option. */
 isc_boolean_t use_if_id = ISC_FALSE;
@@ -139,7 +142,7 @@ static int strip_relay_agent_options(struct interface_info *,
 static void request_v4_interface(const char* name, int flags);
 
 static const char copyright[] =
-"Copyright 2004-2016 Internet Systems Consortium.";
+"Copyright 2004-2018 Internet Systems Consortium.";
 static const char arr[] = "All rights reserved.";
 static const char message[] =
 "Internet Systems Consortium DHCP Relay Agent";
@@ -149,8 +152,30 @@ static const char url[] =
 char *progname;
 
 #ifdef DHCPv6
+#ifdef RELAY_PORT
 #define DHCRELAY_USAGE \
-"Usage: %s [-4] [-d] [-q] [-a] [-D]\n"\
+"Usage: %s [-4] [-d] [-q] [-a] [-D]\n" \
+"                     [-A <length>] [-c <hops>]\n" \
+"                     [-p <port> | -rp <relay-port>]\n" \
+"                     [-pf <pid-file>] [--no-pid]\n"\
+"                     [-m append|replace|forward|discard]\n" \
+"                     [-i interface0 [ ... -i interfaceN]\n" \
+"                     [-iu interface0 [ ... -iu interfaceN]\n" \
+"                     [-id interface0 [ ... -id interfaceN]\n" \
+"                     [-U interface]\n" \
+"                     server0 [ ... serverN]\n\n" \
+"       %s -6   [-d] [-q] [-I] [-c <hops>]\n" \
+"                     [-p <port> | -rp <relay-port>]\n" \
+"                     [-pf <pid-file>] [--no-pid]\n" \
+"                     [-s <subscriber-id>]\n" \
+"                     -l lower0 [ ... -l lowerN]\n" \
+"                     -u upper0 [ ... -u upperN]\n" \
+"           lower (client link): [address%%]interface[#index]\n" \
+"           upper (server link): [address%%]interface\n\n" \
+"       %s {--version|--help|-h}"
+#else
+#define DHCRELAY_USAGE \
+"Usage: %s [-4] [-d] [-q] [-a] [-D]\n" \
 "                     [-A <length>] [-c <hops>] [-p <port>]\n" \
 "                     [-pf <pid-file>] [--no-pid]\n"\
 "                     [-m append|replace|forward|discard]\n" \
@@ -164,8 +189,23 @@ char *progname;
 "                     [-s <subscriber-id>]\n" \
 "                     -l lower0 [ ... -l lowerN]\n" \
 "                     -u upper0 [ ... -u upperN]\n" \
-"       lower (client link): [address%%]interface[#index]\n" \
-"       upper (server link): [address%%]interface"
+"           lower (client link): [address%%]interface[#index]\n" \
+"           upper (server link): [address%%]interface\n\n" \
+"       %s {--version|--help|-h}"
+#endif
+#else /* !DHCPv6 */
+#ifdef RELAY_PORT
+#define DHCRELAY_USAGE \
+"Usage: %s [-d] [-q] [-a] [-D] [-A <length>] [-c <hops>]\n" \
+"                [-p <port> | -rp <relay-port>]\n" \
+"                [-pf <pid-file>] [--no-pid]\n" \
+"                [-m append|replace|forward|discard]\n" \
+"                [-i interface0 [ ... -i interfaceN]\n" \
+"                [-iu interface0 [ ... -iu interfaceN]\n" \
+"                [-id interface0 [ ... -id interfaceN]\n" \
+"                [-U interface]\n" \
+"                server0 [ ... serverN]\n\n" \
+"       %s {--version|--help|-h}"
 #else
 #define DHCRELAY_USAGE \
 "Usage: %s [-d] [-q] [-a] [-D] [-A <length>] [-c <hops>] [-p <port>]\n" \
@@ -175,7 +215,9 @@ char *progname;
 "                [-iu interface0 [ ... -iu interfaceN]\n" \
 "                [-id interface0 [ ... -id interfaceN]\n" \
 "                [-U interface]\n" \
-"                server0 [ ... serverN]\n\n"
+"                server0 [ ... serverN]\n\n" \
+"       %s {--version|--help|-h}"
+#endif
 #endif
 
 /*!
@@ -194,6 +236,12 @@ char *progname;
  * \return Nothing
  */
 static const char use_noarg[] = "No argument for command: %s";
+#ifdef RELAY_PORT
+static const char use_port_defined[] = "Port already set, %s inappropriate";
+#if !defined (USE_BPF_RECEIVE) && !defined (USE_LPF_RECEIVE)
+static const char bpf_sock_support[] = "Only LPF and BPF are supported: %s";
+#endif
+#endif
 #ifdef DHCPv6
 static const char use_badproto[] = "Protocol already set, %s inappropriate";
 static const char use_v4command[] = "Command not used for DHCPv6: %s";
@@ -202,6 +250,10 @@ static const char use_v6command[] = "Command not used for DHCPv4: %s";
 
 static void
 usage(const char *sfmt, const char *sarg) {
+	log_info("%s %s", message, PACKAGE_VERSION);
+	log_info(copyright);
+	log_info(arr);
+	log_info(url);
 
 	/* If desired print out the specific error message */
 #ifdef PRINT_SPECIFIC_CL_ERRORS
@@ -213,6 +265,7 @@ usage(const char *sfmt, const char *sarg) {
 #ifdef DHCPv6
 		  isc_file_basename(progname),
 #endif
+		  isc_file_basename(progname),
 		  isc_file_basename(progname));
 }
 
@@ -223,9 +276,12 @@ main(int argc, char **argv) {
 	struct server_list *sp = NULL;
 	char *service_local = NULL, *service_remote = NULL;
 	u_int16_t port_local = 0, port_remote = 0;
-	int no_daemon = 0, quiet = 0;
+	int quiet = 0;
 	int fd;
 	int i;
+#ifdef RELAY_PORT
+	int port_defined = 0;
+#endif
 #ifdef DHCPv6
 	struct stream_list *sl = NULL;
 	int local_family_set = 0;
@@ -256,9 +312,53 @@ main(int argc, char **argv) {
 	setlogmask(LOG_UPTO(LOG_INFO));
 #endif	
 
+	/* Parse arguments changing no_daemon */
+	for (i = 1; i < argc; i++) {
+		if (!strcmp(argv[i], "-d")) {
+			no_daemon = 1;
+		} else if (!strcmp(argv[i], "--version")) {
+			log_info("isc-dhcrelay-%s", PACKAGE_VERSION);
+			exit(0);
+		} else if (!strcmp(argv[i], "--help") ||
+			   !strcmp(argv[i], "-h")) {
+			log_info(DHCRELAY_USAGE,
+#ifdef DHCPv6
+				 isc_file_basename(progname),
+#endif
+				 isc_file_basename(progname),
+				 isc_file_basename(progname));
+			exit(0);
+		}
+	}
+	/* When not forbidden prepare to become a daemon */
+	if (!no_daemon) {
+		int pid;
+
+		if (pipe(dfd) == -1)
+			log_fatal("Can't get pipe: %m");
+		if ((pid = fork ()) < 0)
+			log_fatal("Can't fork daemon: %m");
+		if (pid != 0) {
+			/* Parent: wait for the child to start */
+			int n;
+
+			(void) close(dfd[1]);
+			do {
+				char buf;
+
+				n = read(dfd[0], &buf, 1);
+				if (n == 1)
+					_exit(0);
+			} while (n == -1 && errno == EINTR);
+			_exit(1);
+		}
+		/* Child */
+		(void) close(dfd[0]);
+	}
+
+
 	/* Set up the isc and dns library managers */
-	status = dhcp_context_create(DHCP_CONTEXT_PRE_DB | DHCP_CONTEXT_POST_DB,
-				     NULL, NULL);
+	status = dhcp_context_create(DHCP_CONTEXT_PRE_DB, NULL, NULL);
 	if (status != ISC_R_SUCCESS)
 		log_fatal("Can't initialize context: %s",
 			  isc_result_totext(status));
@@ -288,16 +388,33 @@ main(int argc, char **argv) {
 			local_family = AF_INET6;
 #endif
 		} else if (!strcmp(argv[i], "-d")) {
-			no_daemon = 1;
+			/* no_daemon = 1; */
 		} else if (!strcmp(argv[i], "-q")) {
 			quiet = 1;
 			quiet_interface_discovery = 1;
 		} else if (!strcmp(argv[i], "-p")) {
 			if (++i == argc)
 				usage(use_noarg, argv[i-1]);
+#ifdef RELAY_PORT
+			if (port_defined)
+				usage(use_port_defined, argv[i-1]);
+			port_defined = 1;
+#endif
 			local_port = validate_port(argv[i]);
 			log_debug("binding to user-specified port %d",
 				  ntohs(local_port));
+#ifdef RELAY_PORT
+		} else if (!strcmp(argv[i], "-rp")) {
+			if (++i == argc)
+				usage(use_noarg, argv[i-1]);
+			if (port_defined)
+				usage(use_port_defined, argv[i-1]);
+			port_defined = 1;
+			relay_port = validate_port(argv[i]);
+			log_debug("binding to user-specified relay port %d",
+				  ntohs(relay_port));
+			add_agent_options = 1;
+#endif
 		} else if (!strcmp(argv[i], "-c")) {
 			int hcount;
 			if (++i == argc)
@@ -481,17 +598,6 @@ main(int argc, char **argv) {
 			no_dhcrelay_pid = ISC_TRUE;
 		} else if (!strcmp(argv[i], "--no-pid")) {
 			no_pid_file = ISC_TRUE;
-		} else if (!strcmp(argv[i], "--version")) {
-			log_info("isc-dhcrelay-%s", PACKAGE_VERSION);
-			exit(0);
-		} else if (!strcmp(argv[i], "--help") ||
-			   !strcmp(argv[i], "-h")) {
-			log_info(DHCRELAY_USAGE,
-#ifdef DHCPv6
-				 isc_file_basename(progname),
-#endif
-				 isc_file_basename(progname));
-			exit(0);
  		} else if (argv[i][0] == '-') {
 			usage("Unknown command: %s", argv[i]);
  		} else {
@@ -529,6 +635,12 @@ main(int argc, char **argv) {
  		}
 	}
 
+#if defined(RELAY_PORT) && \
+    !defined (USE_BPF_RECEIVE) && !defined (USE_LPF_RECEIVE)
+	if (relay_port && (local_family == AF_INET))
+		usage(bpf_sock_support, "-rp");
+#endif
+
 	/*
 	 * If the user didn't specify a pid file directly
 	 * find one from environment variables or defaults
@@ -647,17 +759,21 @@ main(int argc, char **argv) {
 
 	/* Become a daemon... */
 	if (!no_daemon) {
-		int pid;
+		char buf = 0;
 		FILE *pf;
 		int pfdesc;
 
 		log_perror = 0;
 
-		if ((pid = fork()) < 0)
-			log_fatal("Can't fork daemon: %m");
-		else if (pid)
-			exit(0);
+		/* Signal parent we started successfully. */
+		if (dfd[0] != -1 && dfd[1] != -1) {
+			if (write(dfd[1], &buf, 1) != 1)
+				log_fatal("write to parent: %m");
+			(void) close(dfd[1]);
+			dfd[0] = dfd[1] = -1;
+		}
 
+		/* Create the pid file. */
 		if (no_pid_file == ISC_FALSE) {
 			pfdesc = open(path_dhcrelay_pid,
 				      O_CREAT | O_TRUNC | O_WRONLY, 0644);
@@ -1217,6 +1333,12 @@ add_relay_agent_options(struct interface_info *ip, struct dhcp_packet *packet,
 		optlen += 6;
 	}
 
+#ifdef RELAY_PORT
+	if (relay_port) {
+		optlen += 2;
+	}
+#endif
+
 	/* We do not support relay option fragmenting(multiple options to
 	 * support an option data exceeding 255 bytes).
 	 */
@@ -1261,6 +1383,14 @@ add_relay_agent_options(struct interface_info *ip, struct dhcp_packet *packet,
 			log_debug ("Adding link selection suboption"
 				   " with addr: %s", inet_ntoa(giaddr));
 		}
+
+#ifdef RELAY_PORT
+		/* draft-ietf-dhc-relay-port-10.txt section 5.1 */
+		if (relay_port) {
+			*sp++ = RAI_RELAY_PORT;
+			*sp++ = 0u;
+		}
+#endif
 	} else {
 		++agent_option_errors;
 		log_error("No room in packet (used %d of %d) "
@@ -1506,6 +1636,9 @@ setup_streams(void) {
 static const int required_forw_opts[] = {
 	D6O_INTERFACE_ID,
 	D6O_SUBSCRIBER_ID,
+#if defined(RELAY_PORT)
+	D6O_RELAY_SOURCE_PORT,
+#endif
 	D6O_RELAY_MSG,
 	0
 };
@@ -1520,6 +1653,7 @@ process_up6(struct packet *packet, struct stream_list *dp) {
 	struct dhcpv6_relay_packet *relay;
 	struct option_state *opts;
 	struct stream_list *up;
+	u_int16_t relay_client_port = 0;
 
 	/* Check if the message should be relayed to the server. */
 	switch (packet->dhcpv6_msg_type) {
@@ -1580,6 +1714,10 @@ process_up6(struct packet *packet, struct stream_list *dp) {
 			}
 			memset(&relay->link_address, 0, 16);
 		}
+
+		if (packet->client_port != htons(547)) {
+			relay_client_port = packet->client_port;
+		}
 	} else {
 		relay->hop_count = 0;
 		if (!dp)
@@ -1632,6 +1770,30 @@ process_up6(struct packet *packet, struct stream_list *dp) {
 	}
 		
 
+#if defined(RELAY_PORT)
+	/*
+	 * If we use a non-547 UDP source port or if we have received
+	 * from a downstream relay agent uses a non-547 port, we need
+	 * to include the RELAY-SOURCE-PORT option. The "Downstream
+	 * UDP Port" field value in the option allow us to send
+	 * relay-reply message back to the downstream relay agent
+	 * with the correct UDP source port.
+        */
+	if (relay_port || relay_client_port) {
+		if (!save_option_buffer(&dhcpv6_universe, opts, NULL,
+					(unsigned char *) &relay_client_port,
+					sizeof(u_int16_t),
+					D6O_RELAY_SOURCE_PORT, 0)) {
+			log_error("Can't save relay-source-port.");
+			option_state_dereference(&opts, MDL);
+			return;
+		}
+	}
+#else
+	/* Avoid unused but set warning, */
+	(void)(relay_client_port);
+#endif
+
 	/* Add the relay-msg carrying the packet. */
 	if (!save_option_buffer(&dhcpv6_universe, opts,
 				NULL, (unsigned char *) packet->raw,
@@ -1666,6 +1828,9 @@ process_down6(struct packet *packet) {
 	struct data_string relay_msg;
 	const struct dhcpv6_packet *msg;
 	struct data_string if_id;
+#if defined(RELAY_PORT)
+	struct data_string down_port;
+#endif
 	struct sockaddr_in6 to;
 	struct iaddr peer;
 
@@ -1687,6 +1852,9 @@ process_down6(struct packet *packet) {
 	/* Inits. */
 	memset(&relay_msg, 0, sizeof(relay_msg));
 	memset(&if_id, 0, sizeof(if_id));
+#if defined(RELAY_PORT)
+	memset(&down_port, 0, sizeof(down_port));
+#endif
 	memset(&to, 0, sizeof(to));
 	to.sin6_family = AF_INET6;
 #ifdef HAVE_SA_LEN
@@ -1757,6 +1925,37 @@ process_down6(struct packet *packet) {
 		/* Relay-Reply of for another relay, not a client. */
 	      case DHCPV6_RELAY_REPL:
 		to.sin6_port = local_port;
+
+#if defined(RELAY_PORT)
+		oc = lookup_option(&dhcpv6_universe, packet->options,
+				   D6O_RELAY_SOURCE_PORT);
+		if (oc != NULL) {
+			u_int16_t down_relay_port;
+
+			memset(&down_port, 0, sizeof(down_port));
+			if (!evaluate_option_cache(&down_port, packet, NULL,
+						   NULL, packet->options, NULL,
+						   &global_scope, oc, MDL) ||
+			    (down_port.len != sizeof(u_int16_t))) {
+				log_info("Can't evaluate down "
+					 "relay-source-port.");
+				goto cleanup;
+			}
+			memcpy(&down_relay_port, down_port.data,
+			       sizeof(u_int16_t));
+			/*
+			 * If the down_relay_port value is non-zero,
+			 * that means our downstream relay agent uses
+			 * a non-547 UDP source port sending
+			 * relay-forw message to us. We need to use
+			 * the same UDP port sending reply back.
+			 */
+			if (down_relay_port) {
+				to.sin6_port = down_relay_port;
+			}
+		}
+#endif
+
 		/* Fall into: */
 
 	      case DHCPV6_ADVERTISE:
@@ -1847,6 +2046,13 @@ dhcp(struct packet *packet) {
 	return;
 }
 
+#if defined(DHCPv6) && defined(DHCP4o6)
+isc_result_t dhcpv4o6_handler(omapi_object_t *h)
+{
+	return ISC_R_NOTIMPLEMENTED;
+}
+#endif
+
 void
 classify(struct packet *p, struct class *c) {
 	return;
@@ -1870,12 +2076,19 @@ parse_allow_deny(struct option_cache **oc, struct parse *p, int i) {
 isc_result_t
 dhcp_set_control_state(control_object_state_t oldstate,
 		       control_object_state_t newstate) {
+	char buf = 0;
+
 	if (newstate != server_shutdown)
 		return ISC_R_SUCCESS;
 
 	if (no_pid_file == ISC_FALSE)
 		(void) unlink(path_dhcrelay_pid);
 
+	if (!no_daemon && dfd[0] != -1 && dfd[1] != -1) {
+		IGNORE_RET(write(dfd[1], &buf, 1));
+		(void) close(dfd[1]);
+		dfd[0] = dfd[1] = -1;
+	}
 	exit(0);
 }
 
